"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const prettier_1 = require("prettier");
const parser_typescript_1 = require("prettier/parser-typescript");
const package_phrases_1 = require("./metadata/package-phrases");
const jsplugin_1 = require("./printer/jsplugin");
const multiline_array_printer_1 = require("./printer/multiline-array-printer");
const multilineArraysPlugin = {
    languages: [
        {
            name: 'typescript',
            parsers: [
                package_phrases_1.parserName,
            ],
        },
    ],
    parsers: {
        [package_phrases_1.parserName]: {
            ...parser_typescript_1.parsers['typescript'],
            astFormat: package_phrases_1.astFormatName,
        },
    },
    printers: {
        [package_phrases_1.astFormatName]: multiline_array_printer_1.multilineArrayPrinter,
    },
};
function addMultilinePrinter(options) {
    var _a;
    if ('printer' in options) {
        (0, jsplugin_1.setOriginalPrinter)(options.printer);
        // overwrite the printer with ours
        options.printer = multiline_array_printer_1.multilineArrayPrinter;
    }
    else {
        /** If the printer hasn't already been assigned in options, rearrange plugins so that ours gets chosen. */
        const plugins = options.plugins;
        const firstMatchedPlugin = plugins.find((plugin) => typeof plugin !== 'string' && !!plugin.printers && !!plugin.printers.estree);
        if (!firstMatchedPlugin || typeof firstMatchedPlugin === 'string') {
            throw new Error(`Matched invalid first plugin: ${firstMatchedPlugin}`);
        }
        const matchedPrinter = (_a = firstMatchedPlugin.printers) === null || _a === void 0 ? void 0 : _a.estree;
        if (!matchedPrinter) {
            throw new Error(`Printer not found on matched plugin: ${firstMatchedPlugin}`);
        }
        (0, jsplugin_1.setOriginalPrinter)(matchedPrinter);
        const thisPluginIndex = plugins.findIndex((plugin) => {
            return plugin.pluginMarker === pluginMarker;
        });
        const thisPlugin = plugins[thisPluginIndex];
        if (!thisPlugin) {
            throw new Error(`This plugin was not found.`);
        }
        // remove this plugin from its current location in the array
        plugins.splice(thisPluginIndex, 1);
        // add this plugin to the beginning of the array so its printer is found first
        plugins.splice(0, 0, thisPlugin);
    }
}
const languages = (0, prettier_1.getSupportInfo)().languages.filter(({ name }) => [
    'JavaScript',
    'Flow',
    'JSX',
    'TSX',
    'TypeScript',
    'Markdown',
    'MDX',
].includes(name));
function findPluginsByParserName(parserName, options) {
    return options.plugins.filter((plugin) => {
        var _a;
        return (typeof plugin === 'object' &&
            plugin.pluginMarker !== pluginMarker &&
            !!((_a = plugin.parsers) === null || _a === void 0 ? void 0 : _a[parserName]));
    });
}
function mergeParsers(originalParser, parserName) {
    const thisPluginPreprocess = (text, options) => {
        const pluginsWithPreprocessor = findPluginsByParserName(parserName, options).filter((plugin) => { var _a, _b; return !!((_b = (_a = plugin.parsers) === null || _a === void 0 ? void 0 : _a[parserName]) === null || _b === void 0 ? void 0 : _b.preprocess); });
        let processedText = text;
        pluginsWithPreprocessor.forEach((pluginWithPreprocessor) => {
            var _a, _b, _c;
            const nextText = (_c = (_b = (_a = pluginWithPreprocessor.parsers) === null || _a === void 0 ? void 0 : _a[parserName]) === null || _b === void 0 ? void 0 : _b.preprocess) === null || _c === void 0 ? void 0 : _c.call(_b, processedText, {
                ...options,
                plugins: options.plugins.filter((plugin) => plugin.pluginMarker !== pluginMarker),
            });
            if (nextText != undefined) {
                processedText = nextText;
            }
        });
        addMultilinePrinter(options);
        // Object.assign(parser, {
        //     ...parser,
        //     preprocess: thisPluginPreprocess,
        // });
        return processedText;
    };
    const parser = {
        ...originalParser,
        preprocess: thisPluginPreprocess,
    };
    return parser;
}
const typescriptParser = mergeParsers(parser_typescript_1.parsers.typescript, 'typescript');
const pluginMarker = {};
const plugin = {
    // used to ensuring tracking of this plugin later
    ...{ pluginMarker: pluginMarker },
    languages,
    parsers: {
        typescript: typescriptParser,
    },
    printers: {
        estree: multiline_array_printer_1.multilineArrayPrinter,
    },
};
module.exports = plugin;
