"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.printWithMultilineArrays = void 0;
const prettier_1 = require("prettier");
const package_phrases_1 = require("../metadata/package-phrases");
const child_docs_1 = require("./child-docs");
const nestingSyntaxOpen = '[{(<';
const nestingSyntaxClose = ']})>';
function isDocCommand(inputDoc) {
    return !!inputDoc && typeof inputDoc !== 'string' && !Array.isArray(inputDoc);
}
const found = 'Found "[" but';
function insertArrayLines(inputDoc, lineCounts, debug) {
    (0, child_docs_1.walkDoc)(inputDoc, debug, (currentDoc, parentDocs, childIndex) => {
        const currentParent = parentDocs[0];
        const parentDoc = currentParent === null || currentParent === void 0 ? void 0 : currentParent.parent;
        if (typeof currentDoc === 'string' && currentDoc.trim() === '[') {
            if (debug) {
                console.info({ currentDoc, parentDoc });
            }
            if (!Array.isArray(parentDoc) || parentDoc.length !== 4) {
                throw new Error(`${found} parentDoc is not an array.`);
            }
            if (childIndex !== 0) {
                throw new Error(`${found} not at index 0 in its parent`);
            }
            const bracketSibling = parentDoc[childIndex + 1];
            if (debug) {
                console.info({ bracketSibling });
            }
            if (bracketSibling === ']') {
                return false;
            }
            if (!isDocCommand(bracketSibling) || bracketSibling.type !== 'indent') {
                throw new Error(`${found} its sibling was not an indent Doc.: ${bracketSibling}`);
            }
            const indentContents = bracketSibling.contents;
            if (debug) {
                console.info({ indentContents });
            }
            if (!Array.isArray(indentContents)) {
                throw new Error(`${found} indent didn't have array contents.`);
            }
            if (indentContents.length !== 3) {
                throw new Error(`${found} indent contents did not have 3 children.`);
            }
            const startingLine = indentContents[0];
            if (debug) {
                console.info({ firstIndentContentsChild: startingLine });
            }
            if (!isDocCommand(startingLine) || startingLine.type !== 'line') {
                throw new Error(`${found} first indent child was not a line.`);
            }
            indentContents[0] = '';
            const indentedContent = indentContents[1];
            if (debug) {
                console.info({
                    secondIndentContentsChild: indentedContent,
                    itsFirstChild: indentedContent[0],
                });
            }
            if (!indentedContent) {
                console.error('second indent child (indentedContent) is not defined:', {
                    indentContents,
                    indentedContent,
                });
                throw new Error(`${found} second indent child is not a fill.`);
            }
            if (!Array.isArray(indentedContent) &&
                !(isDocCommand(indentedContent) && indentedContent.type === 'fill')) {
                console.error('second indent child (indentCode) is not a fill doc or an array:', {
                    indentContents,
                    indentCode: indentedContent,
                });
                throw new Error(`${found} second indent child is not a fill doc or an array.`);
            }
            if (Array.isArray(indentedContent)
                ? indentedContent.length === 0
                : indentedContent.parts.length === 0) {
                throw new Error(`${found} indentedContent has no length.`);
            }
            // lineIndex is 0 indexed
            let lineIndex = 0;
            // columnCount is 1 indexed
            let columnCount = 1;
            if (debug) {
                console.info(`>>>>>>>>>>>>>> Walking children for commas`);
            }
            (0, child_docs_1.walkDoc)(indentedContent, debug, (currentDoc, commaParents, commaChildIndex) => {
                const innerCurrentParent = commaParents[0];
                const commaParentDoc = innerCurrentParent === null || innerCurrentParent === void 0 ? void 0 : innerCurrentParent.parent;
                if (isDocCommand(currentDoc) && currentDoc.type === 'if-break') {
                    if (!commaParentDoc) {
                        throw new Error(`Found if-break without a parent`);
                    }
                    if (!Array.isArray(commaParentDoc)) {
                        throw new Error(`if-break parent is not an array`);
                    }
                    if (commaChildIndex == undefined) {
                        throw new Error(`if-break child index is undefined`);
                    }
                    commaParentDoc[commaChildIndex] = currentDoc.breakContents;
                    commaParentDoc.splice(commaChildIndex + 1, 0, prettier_1.doc.builders.breakParent);
                }
                else if (typeof currentDoc === 'string') {
                    if (nestingSyntaxOpen.includes(currentDoc)) {
                        return false;
                    }
                    else if (nestingSyntaxClose.includes(currentDoc)) {
                        throw new Error(`Found closing syntax which shouldn't be walked`);
                    }
                    else if (currentDoc === ',') {
                        if (debug) {
                            console.info({ foundCommaIn: commaParentDoc });
                        }
                        if (!commaParentDoc) {
                            console.error({ innerParentDoc: commaParentDoc });
                            throw new Error(`Found comma but innerParentDoc is not defined.`);
                        }
                        if (!Array.isArray(commaParentDoc)) {
                            console.error({ innerParentDoc: commaParentDoc });
                            throw new Error(`Found comma but innerParentDoc is not an array.`);
                        }
                        if (commaChildIndex == undefined) {
                            throw new Error(`Found comma but childIndex is undefined.`);
                        }
                        let siblingIndex = commaChildIndex + 1;
                        let parentToMutate = commaParentDoc;
                        if (commaChildIndex === commaParentDoc.length - 1) {
                            const commaGrandParent = commaParents[1];
                            if (commaGrandParent == undefined) {
                                throw new Error(`Could not find grandparent of comma group.`);
                            }
                            if (commaGrandParent.childIndexInThisParent == undefined) {
                                throw new Error(`Could not find index of comma group parent`);
                            }
                            if (!Array.isArray(commaGrandParent.parent)) {
                                throw new Error(`Comma group grandparent is not an array.`);
                            }
                            siblingIndex = commaGrandParent.childIndexInThisParent + 1;
                            parentToMutate = commaGrandParent.parent;
                        }
                        if (debug) {
                            console.info({ commaParentDoc, parentToMutate, siblingIndex });
                        }
                        const commaSibling = parentToMutate[siblingIndex];
                        if (!isDocCommand(commaSibling) || commaSibling.type !== 'line') {
                            throw new Error(`Found comma but its following sibling is not a line: ${commaSibling}`);
                        }
                        const currentLineCountIndex = lineIndex % lineCounts.length;
                        const currentLineCount = lineCounts.length
                            ? lineCounts[currentLineCountIndex]
                            : undefined;
                        if ((currentLineCount && columnCount === currentLineCount) ||
                            !currentLineCount) {
                            // if we're on the last element of the line then increment to the next line
                            lineIndex++;
                            columnCount = 1;
                            /**
                             * Don't use doc.builders.hardline here. It causes "invalid size
                             * error" which I don't understand and which has no other useful
                             * information or stack trace.
                             */
                            if (debug) {
                                console.info({ breakingAfter: parentToMutate[siblingIndex - 1] });
                            }
                            parentToMutate[siblingIndex] =
                                prettier_1.doc.builders.hardlineWithoutBreakParent;
                        }
                        else {
                            parentToMutate[siblingIndex] = ' ';
                            columnCount++;
                        }
                    }
                }
                return true;
            });
            if (Array.isArray(indentedContent)) {
                indentContents.splice(1, 1, prettier_1.doc.builders.group([
                    prettier_1.doc.builders.hardlineWithoutBreakParent,
                    ...indentedContent,
                ]));
            }
            else {
                indentedContent.parts = [
                    prettier_1.doc.builders.group([
                        prettier_1.doc.builders.hardlineWithoutBreakParent,
                        ...indentedContent.parts,
                    ]),
                ];
            }
            // don't walk any deeper
            return false;
        }
        else if (debug) {
            console.info({ ignoring: currentDoc });
        }
        return true;
    });
    // return what is input because we perform mutations on it
    return inputDoc;
}
const ignoreTheseKeys = [
    'tokens',
];
const ignoreTheseChildTypes = [
    'string',
    'number',
];
function isMaybeComment(input) {
    if (!input || typeof input !== 'object') {
        return false;
    }
    if (!('type' in input)) {
        return false;
    }
    if (input.type.toLowerCase() !== 'block' && input.type.toLowerCase() !== 'line') {
        return false;
    }
    if (!('value' in input)) {
        return false;
    }
    return true;
}
function extractComments(node) {
    debugger;
    if (!node || typeof node !== 'object') {
        return [];
    }
    let comments = [];
    if (Array.isArray(node)) {
        comments.push(...node.filter(isMaybeComment));
    }
    Object.keys(node).forEach((nodeKey) => {
        if (!ignoreTheseKeys.includes(nodeKey)) {
            const nodeChild = node[nodeKey];
            if (!ignoreTheseChildTypes.includes(typeof nodeChild)) {
                comments.push(...extractComments(nodeChild));
            }
        }
    });
    // this might duplicate comments but our later code doesn't care
    return comments;
}
const mappedComments = new WeakMap();
function isProgramNode(input) {
    return input.type === 'Program';
}
function setLineCounts(programNode, debug) {
    // parse comments only on the program node so it only happens once
    const comments = extractComments(programNode);
    if (debug) {
        console.info({ comments });
    }
    const keyedCommentsByLastLine = comments.reduce((accum, currentComment) => {
        var _a;
        const commentText = (_a = currentComment.value) === null || _a === void 0 ? void 0 : _a.replace(/\n/g, ' ');
        if (commentText === null || commentText === void 0 ? void 0 : commentText.toLowerCase().includes(package_phrases_1.elementsPerLineTrigger)) {
            const split = commentText
                .toLowerCase()
                .replace(package_phrases_1.untilLineTriggerRegExp, '')
                .replace(/,/g, '')
                .split(' ');
            const numbers = split
                .map((entry) => (entry && entry.trim().match(/\d+/) ? Number(entry) : NaN))
                .filter((entry) => !isNaN(entry));
            if (!currentComment.loc) {
                throw new Error(`Cannot read line location for comment ${currentComment.value}`);
            }
            accum[currentComment.loc.end.line] = numbers;
        }
        return accum;
    }, {});
    // save to a map so we don't have to recalculate these every time
    mappedComments.set(programNode, keyedCommentsByLastLine);
    return keyedCommentsByLastLine;
}
function printWithMultilineArrays(originalFormattedOutput, path, debug) {
    var _a, _b;
    const programNode = path.stack[0];
    if (!isProgramNode(programNode)) {
        throw new Error(`Could not find valid program node in ${path.stack.join(',')}`);
    }
    const node = path.getValue();
    if (node) {
        if (node.type === 'ArrayExpression' ||
            node.type === 'ArrayPattern' ||
            node.type === 'TupleExpression') {
            if (!node.loc) {
                throw new Error(`Could not find location of node ${'value' in node ? node.value : node.type}`);
            }
            const lastLine = node.loc.start.line - 1;
            const lineCountMap = (_a = mappedComments.get(programNode)) !== null && _a !== void 0 ? _a : setLineCounts(programNode, debug);
            const lineCounts = (_b = lineCountMap[lastLine]) !== null && _b !== void 0 ? _b : [];
            if (debug) {
                console.info(`======= Starting call to ${insertArrayLines.name}: =======`);
                console.info({ lineCounts });
            }
            const newDoc = insertArrayLines(originalFormattedOutput, lineCounts, debug);
            return newDoc;
        }
    }
    return originalFormattedOutput;
}
exports.printWithMultilineArrays = printWithMultilineArrays;
